import threading 
import time
import os
import sys
from . import part, multipart_collector, cookie, secured_cookie_value
from . import wsgi_executor, xmlrpc_executor
from skitai.lib import producers
from skitai.server import utility
from hashlib import md5
import random
import base64
from skitai.saddle import cookie
	
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib

multipart_collector.MultipartCollector.file_max_size = 20 * 1024 * 1024
multipart_collector.MultipartCollector.cache_max_size = 5 * 1024 * 1024
secured_cookie_value.Session.default_session_timeout = 1200

class Config:
	pass
	
		
class Saddle (part.Part):
	use_reloader = False
	debug = False
	
	# Session
	securekey = None
	session_timeout = None
	
	#WWW-Authenticate
	authorization = "digest"
	realm = None
	user = None
	password = None
	
	opaque = None
	
	def __init__ (self, module_name, mount = "/"):
		part.Part.__init__ (self, module_name)		
		self.lock = threading.RLock ()
		self.cache_sorted = 0
		self.cached_paths = {}
		self.cached_rules = []
		self.config = Config ()
		
	def __setattr__ (self, name, attr):
		if name == "upload_file_max_size":
			multipart_collector.MultipartCollector.file_max_size = attr
		self.__dict__ [name] = attr
		
	def get_www_authenticate (self):
		if self.authorization == "basic":
			return 'Basic realm="%s"' % self.realm
		else:	
			if self.opaque is None:
				self.opaque = md5 (self.realm.encode ("utf8")).hexdigest ()
			return 'Digest realm="%s", qop="auth", nonce="%s", opaque="%s"' % (
				self.realm, utility.md5uniqid (), self.opaque
			)
			
	def authorize (self, auth, method, uri):
		if self.realm is None or self.user is None or self.password is None:
			return		
		if auth is None:
			return self.get_www_authenticate ()
		
		# check validate: https://evertpot.com/223/
		amethod, authinfo = auth.split (" ", 1)
		if amethod.lower () != self.authorization:
			return self.get_www_authenticate ()
			
		if self.authorization == "basic":
			basic = base64.decodestring (authinfo.encode ("utf8"))
			if basic == "%s:%s" % (self.username, self.password):
				return
				
		else:
			method = method.upper ()
			infod = {}
			for info in authinfo.split (","):
				k, v = info.strip ().split ("=", 1)
				if not v: return self.get_www_authenticate ()
				if v[0] == '"': v = v [1:-1]
				infod [k]	 = v
			
			try:
				if uri != infod ["uri"]:					
					return self.get_www_authenticate ()
				A1 = md5 (("%s:%s:%s" % (self.user, self.realm, self.password)).encode ("utf8")).hexdigest ()
				A2 = md5 (("%s:%s" % (method, infod ["uri"])).encode ("utf8")).hexdigest ()
				Hash = md5 (("%s:%s:%s:%s:%s:%s" % (
					A1, 
					infod ["nonce"],
					infod ["nc"],
					infod ["cnonce"],
					infod ["qop"],
					A2
					)).encode ("utf8")
				).hexdigest ()

				if Hash == infod ["response"]:				
					return
					
			except KeyError:
				pass
		
		return self.get_www_authenticate ()
			
	def set_devel (self, debug = True, use_reloader = True):
		self.debug = debug
		self.use_reloader = use_reloader
	
	def get_multipart_collector (self):
		return multipart_collector.MultipartCollector
	
	def get_method (self, path_info):
		method, kargs = None, {}
		self.lock.acquire ()
		try:
			method = self.cached_paths [path_info]
		except KeyError:
			for rulepack, freg in self.cached_rules:
				method, kargs = self.try_rule (path_info, rulepack)
				if method: 
					break
		finally:
			self.lock.release ()
	
		if not method:
			if self.use_reloader:
				self.lock.acquire ()																
			try:	
				current_app, method, kargs, match, matchtype = self.get_package_method (path_info, self.use_reloader)
			finally:	
				if self.use_reloader: 
					self.lock.release ()
			
			if not self.use_reloader:
				self.lock.acquire ()
				if matchtype == 1:
					self.cached_paths [match] = method				
				elif matchtype == 2:
					self.cached_rules.append ([match, 1])
					if time.time () - self.cache_sorted > 300:
						self.cached_rules.sort (lambda x, y: cmp (y[1], x[1]))
						self.cached_rules.sort (key = lambda x: x[1], reverse = True)
						self.cache_sorted = time.time ()
				self.lock.release ()
					
			if matchtype == 3:
				return method, 301
		
		return current_app, method, kargs
	
	def restart (self, wasc, route):
		self.wasc = wasc
		self.route = route
		if self._onreload:
			self._onreload (self.wasc, self)		
	
	def create_on_demand (self, was, name):
		class G: pass
		# create just in time objects
		if name == "cookie":			
			return cookie.Cookie (was.request, self.securekey, self.session_timeout)
			
		if name in ("session", "mbox"):			
			if not was.in__dict__ ("cookie"):
				was.cookie = cookie.Cookie (was.request, self.securekey, self.session_timeout)			
			if name == "session":
				return was.cookie.get_session ()					
			if name == "mbox":
				return was.cookie.get_notices ()		
				
		if name == "g":
			return G ()
	
	def cleanup_on_demands (self, was):
		if was.in__dict__ ("g"):
			del was.g
		if not was.in__dict__ ("cookie"):
			return
		for j in ("session", "mbox"):
			if was.in__dict__ (j):		
				delattr (was, j)
		del was.cookie
										
	def __call__ (self, env, start_response):
		was = env ["skitai.was"]		
		was.app = self
		was.ab = self.build_url
		was.response = was.request.response
		
		content_type = env.get ("CONTENT_TYPE", "")				
		if content_type.startswith ("text/xml") or content_type.startswith ("application/xml+rpc"):
			result = xmlrpc_executor.Executor (env, self.get_method) ()
		else:	
			result = wsgi_executor.Executor (env, self.get_method) ()		
		
		del was.response
		del was.ab		
		self.cleanup_on_demands (was) # del session, mbox, cookie, g
			
		return result
		


#!/usr/bin/python

from skitai.saddle import Saddle
import shutil, os

app = Saddle (__name__)
app.debug = True
app.use_reloader = True

app.securekey = "iodfjksdfkjsdhkfjsd0987987sdf"
app.realm = "Skitai API"
app.user = "app"
app.password = "1111"
app.authorization = "digest"

MULTIPART = """
<form action = "/" enctype="multipart/form-data" method="post">
	<input type="hidden" name="submit-hidden" value="Genious">   
	<p></p>What is your name? <input type="text" name="submit-name" value="Hans Roh"></p>
	<p></p>What files are you sending? <br />
	<input type="file" name="file1"><br />
	<input type="file" name="file2">
	</p>
	<input type="submit" value="Send"> 
	<input type="reset">
</form>
"""

FORMDATA = """
<form action = "/" method="post">
	<input type="hidden" name="submit-hidden" value="Genious">   
	<p></p>What is your name? <input type="text" name="submit-name" value="Hans Roh"></p>	
	<input type="submit" value="Send"> 
	<input type="reset">
</form>
"""
import skitaipackage

app.add_package (skitaipackage, "package")

@app.route ("/fancy/<int:cid>/<cname>")
def fancy (was, cid, cname, zipcode):
	return [
	"%s - %s (%s)" % (cid, cname, zipcode),
	"<hr />",
	was.ab ("fancy", 200, "Skitai Inc", "31052")
	]

@app.route('/')
def hello_world (was, **form):
	if was.request.command == 'post':		
		file = form.get ("file1")
		if file:
			file.save ("d:\\var\\upload", dup = "o")
		return str (form)
	return [was.ab ("fancy", 200, "Skitai Inc", "31052"), FORMDATA, "<hr />", MULTIPART]

@app.route('/indians')
def hello_world (was, num = 8):
	if was.request.command == 'get':		
		was.response ["Content-Type"] = "text/xml"
		return was.toxml ((num,))
	else:
		return num	

@app.route('/ping')
def ping (was, **form):
	return "pong"

	
if __name__ == '__main__':
    app.run(host="0.0.0.0", port=5002)




import sqlite3
from . import dbconnect
import threading

class OperationalError (Exception):
	pass


class SynConnect (dbconnect.DBConnect):
	def __init__ (self, address, params = None, lock = None, logger = None):
		dbconnect.DBConnect.__init__ (self, address, params, lock, logger)
		self.connected = False
	
	def connect (self):
		try:
			self.conn = sqlite3.connect (self.address, check_same_thread = False)
			self.cur = self.conn.cursor ()
		except:
			self.handle_error ()
		else:	
			self.connected = True
	
	def close (self):
		self.connected = False
		dbconnect.DBConnect.close (self)
	
	def execute (self, sql, callback):
		self.callback = callback
		self.has_result = False
		self.exception_str = ""
		self.exception_class = None		
		self.set_event_time ()		
		self.execute_count += 1		
		
		if not self.connected:
			self.connect ()

		sql= sql.strip ()
		try:
			if sql [:7].lower () == "select ":
				self.cur.execute (sql)
			else:	
				self.cur.executescript (sql)				
		except:
			self.handle_error ()
		else:
			self.conn.commit ()
			self.has_result = True		
			self.close_case ()
		


from . import eurl
from . import localstorage
from skitai.protocol.http import request as http_request
from skitai.protocol.http import response as http_response
from skitai.protocol.http import request_handler as http_request_handler
from skitai.protocol.http import tunnel_handler
from skitai.protocol.ws import request_handler as ws_request_handler
from skitai.protocol.ws import tunnel_handler as ws_tunnel_handler
from skitai.protocol.ws import request as ws_request
from skitai.protocol.dns import asyndns
from skitai.client import socketpool, asynconnect
from skitai.server.threads import trigger
from skitai.lib import strutil
from skitai import lifetime
import asyncore
from . import rc
import asyncore
from skitai.client import adns

_currents = {}
_latest = ""
_map = asyncore.socket_map
_logger = None
_debug = False
_que = []
_max_numpool = 4
_current_numpool = 4
_concurrents = 2
_default_header = ""
_use_lifetime = True
_timeout = 30

def configure (
	logger = None, 
	numpool = 3,
	timeout = 30,
	concurrents = 2,
	default_option = "", 
	response_max_size = 100000000,
	use_lifetime = True,
	debug = False
	):
	
	global _logger, _max_numpool, _current_numpool, _concurrents, _default_option, _configured, _use_lifetime, _timeout, _debug
	
	_default_option = default_option
	_max_numpool = numpool + 1
	_current_numpool = _max_numpool
	_logger = logger
	_concurrents = concurrents
	_use_lifetime = use_lifetime
	_debug = debug
	
	http_response.Response.SIZE_LIMIT = response_max_size
	localstorage.create (logger)
		
	socketpool.create (logger)
	adns.init (logger)
	trigger.start_trigger () # for keeping lifetime loop
			
	
def add (thing, callback, front = False):
	global _que, _default_header, _logger, _current_numpool, _currents
	
	if strutil.is_str_like (thing):
		thing = thing + " " + _default_option
		try:
			thing = eurl.EURL (thing)
		except:
			_logger.trace ()
			return
	
	if front:
		_que.insert (0, (thing, callback, _logger))
	else:	
		_que.append ((thing, callback, _logger))
		
	# notify new item
	if thing ["netloc"] not in _currents:
		_current_numpool += 1
	
	if _latest: # after loop started, not during queueing
		maybe_pop ()

def maybe_pop ():
	global _max_numpool, _current_numpool, _que, _map, _concurrents
	global _logger, _use_lifetime, _debug, _currents, _latest
	
	lm = len (_map)
	if _use_lifetime and not _que and lm == 1:
		lifetime.shutdown (0, 1)
		return
	
	if _current_numpool > _max_numpool:
		_current_numpool = _max_numpool  # maximum
	
	_currents = {}
	for r in list (_map.values ()):
		if isinstance (r, asynconnect.AsynConnect) and r.handler: 
			netloc = r.handler.request.el ["netloc"]			
		elif isinstance (r, asyndns.async_dns): 
			netloc = r.qname						
		else:
			continue	
			
		try: _currents [netloc] += 1
		except KeyError: _currents [netloc] = 1
		
	index = 0
	indexes = []		
	while lm < _current_numpool and _que:
		try:
			item = _que [index]
		except IndexError:
			# for minimize cost to search new item by concurrents limitation
			_current_numpool = len (_currents) * _concurrents
			if _current_numpool < 2:
				_current_numpool = 2 # minmum	
			if _debug:
				print (">>>>>>>>>>>> resize numpool %d" % _current_numpool)
			break
		
		el = item [0]
		if _currents.get (el ["netloc"], 0) < _concurrents:
			try: 
				_currents [el ["netloc"]] += 1
			except KeyError:
				_currents [el ["netloc"]] = 1
			indexes.append (index)
			lm += 1
		index += 1
	
	if _debug:
		print (">>>>>>>>>>>> numpool:%d in map:%d queue:%d" % (_current_numpool, len (_map), len (_que)))
	
	pup = 0
	for index in indexes:
		item = _que.pop (index - pup)		
		Item (*item)		
		pup += 1

	if pup:
		# for multi threading mode
		trigger.the_trigger.pull_trigger ()
	
	return pup	

def get_all ():
	import time
	global _use_lifetime, _map, _que, _logger, _concurrent
	
	if not _que:
		_logger ("[warn] no item to get")
		return
	
	for i in range (_current_numpool):
		if not maybe_pop ():
			break
	
	if not _use_lifetime: 
		return
	
	try:
		lifetime.loop (3.0)
	finally:	
		socketpool.cleanup ()


class SSLProxyTunnelHandler (tunnel_handler.SSLProxyTunnelHandler):
	def get_handshaking_buffer (self):	
		req = ("CONNECT %s:%d HTTP/%s\r\nUser-Agent: %s\r\n\r\n" % (
					self.request.el ["netloc"], 
					self.request.el ["port"],
					self.request.el ["http-version"],
					self.request.el.get_useragent ()
				)).encode ("utf8")
		return req
		
class WSSSLProxyTunnelHandler (SSLProxyTunnelHandler, ws_tunnel_handler.SSLProxyTunnelHandler):
	pass

class WSProxyTunnelHandler (ws_tunnel_handler.ProxyTunnelHandler, SSLProxyTunnelHandler):
	def get_handshaking_buffer (self):	
		return SSLProxyTunnelHandler.get_handshaking_buffer (self)


class HTTPRequest (http_request.HTTPRequest):
	def __init__ (self, el, logger = None):		
		self.el = el			
		url = self.el ["rfc"]
		method = self.el ["method"].upper ()
		http_request.HTTPRequest.__init__ (self, url, method, headers = self.el.get_header (), logger = logger)
			
	def split (self, uri):
		return (self.el ["netloc"], self.el ["port"]), self.el ["uri"]
		
	def serialize (self):
		pass
		
	def get_auth (self):
		return self.el ["http-auth"]
			
	def get_data (self):
		if self.el ["scheme"] in ("ws", "wss"): return b""
		return self.el ["http-form"] is not None and self.el ["http-form"].encode ("utf8") or b""

	def get_eurl (self):
		return self.el
	
	def get_useragent (self):
		return self.el ["http-user-agent"]	
		

class WSRequest (HTTPRequest, ws_request.Request):
	def __init__ (self, el, logger = None):
		self.el = el
		ws_request.Request.__init__ (self, self.el ["rfc"], self.el ["wsoc-message"], self.el ["wsoc-opcode"], self.el ["http-auth"], logger = logger)
		
		
class Item:
	def __init__ (self, thing, callback, logger = None):
		global _logger, _timeout
		
		if localstorage.localstorage is None:
			configure (logger)
		
		if strutil.is_str_like (thing):
			self.el = eurl.EURL (thing)
		else:
			self.el = thing
			
		if logger:
			self.logger = logger			
		else:
			self.logger = _logger
		self.callback = callback
		
		if self.el ["scheme"] in ("ws", "wss"):
			self.el ['http-connection'] = "keep-aluve, Upgrade"			
			self.el.to_version_11 ()
			request = WSRequest (self.el, logger = self.logger)
			# websocket proxy should be tunnel
			if self.el.has_key ("http-proxy"):
				self.el ["http-tunnel"] = self.el ["http-proxy"]
				del self.el ["http-proxy"]
			if self.el.has_key ("http-tunnel"):
				if self.el ['scheme'] == 'wss':
					handler_class = WSSSLProxyTunnelHandler				
				else:
					handler_class = WSProxyTunnelHandler		
			else:
				handler_class = ws_request_handler.RequestHandler
				
		else:
			request = HTTPRequest (self.el, logger = self.logger)
			if self.el.has_key ("http-tunnel"):		
				request.el.to_version_11 ()
				handler_class = SSLProxyTunnelHandler
			else:	
				handler_class = http_request_handler.RequestHandler
			
		sp = socketpool.socketpool		
		if request.el ["http-tunnel"]:
			asyncon = sp.get ("proxys://%s" % request.el ["http-tunnel"])
		elif request.el ["http-proxy"]:
			asyncon = sp.get ("proxy://%s" % request.el ["http-proxy"])
		else:
			asyncon = sp.get (request.el ["rfc"])
					
		handler_class (asyncon, request, self.callback_wrap, request.el ["http-version"], request.el ['connection']).start ()
	
	def handle_websocket (self, handler):
		if handler.response.code == 101:
			request = WSRequest (handler.request.el, logger = self.logger)
			ws_request_handler.RequestHandler (asyncon, request, self.callback_wrap).start ()
		
	def callback_wrap (self, handler):
		global _latest
	
		r = rc.ResponseContainer (handler, self.callback)		
		_latest = r.uinfo.netloc
		# unkink back refs
		handler.asyncon = None
		handler.callback = None
		handler.response = None
		handler.request = None
		del handler
		
		try:
			self.callback (r)		
		except:
			_logger.trace ()
		maybe_pop ()



import pickle as pickle
from skitai.server import http_date
import time
import random
from skitai.lib import pathtool
from skitai.lib import udict
import os, sys
try:
	from urllib.parse import quote_plus, unquote_plus
except ImportError:
	from urllib import quote_plus, unquote_plus	
import base64
import pickle
from hmac import new as hmac

_default_hash = None
if sys.version_info >= (2, 5):
	try:
		from hashlib import sha1 as _default_hash
	except ImportError:
		pass
if _default_hash is None:
	import sha as _default_hash


class UnquoteError(Exception):
	"""Internal exception used to signal failures on quoting."""


class SecuredValue:
	hash_method = _default_hash
	serialization_method = pickle
	quote_base64 = True
	KEY = "SECURED"	
	
	def __init__ (self, request, data, secret_key, setfunc, new = True):
		self.request = request
		if new:
			self.set_default_data ()
		else:
			self.data = data				
		self.secret_key = secret_key
		self.setfunc = setfunc
		self.new = new
		self.dirty = False
		
	def __contains__ (self, k):
		return k in self.data
		
	def set_default_data (self):
		self.dirty = False
		self.data = None
	
	def clear (self):
		self.set_default_data ()
		
	def recal_expires (self, expires):						
		return expires
	
	def rollback (self):
		self.dirty = False
			
	def commit (self, expires = None, path = "/", domain = None):
		if not self.dirty:
			return
		expires = self.recal_expires (expires)
		self.setfunc (self.KEY, self.serialize (), expires, path, domain)
		self.dirty = False
		
	@classmethod
	def quote (cls, value):		
		if cls.serialization_method is not None:
			value = cls.serialization_method.dumps(value, 1)			
		if cls.quote_base64:
			value = base64.b64encode (value)
			value = b''.join(value.splitlines()).strip()
		return value
	
	@classmethod
	def unquote(cls, value):
		try:
			if cls.quote_base64:
				value = base64.b64decode(value)
			if cls.serialization_method is not None:
				value = cls.serialization_method.loads(value)
			return value
		except:			
			raise UnquoteError


class SecuredDictValue (SecuredValue):
	KEY = "SECUREDICT"	
			
	def set_default_data (self):
		self.dirty = True
		self.data = {}
			
	def __setitem__ (self, k, v):
		self.set (k, v)
	
	def __delitem__ (self, k):
		return self.remove (k)
	
	def __getitem__ (self, k, v = None):
		return self.data.get (k, v)
	
	def iterkeys (self):		
		return self.data.iterkeys ()
	
	def itervalues (self):		
		return self.data.itervalues ()	
	
	def iteritems (self):		
		return self.data.iteritems ()	
		
	def has_key (self, k):
		return k in self.data
	
	def items (self):
		return list(self.data.items ())
	
	def keys (self):
		return list(self.data.keys ())	
	
	def values (self):
		return list(self.data.values ())	
			
	def remove (self, k):
		try:
			del self.data [k]
		except KeyError:
			pass
		else:
			self.dirty = True
			
	def set (self, k, v):
		if type (k) is not type (""):
			raise TypeError("Session key must be string type")
		self.data [k] = v
		self.dirty = True
	
	def get (self, k, v = None):
		return self.data.get (k, v)
		
	def serialize(self):
		if self.secret_key is None:
			raise RuntimeError('no secret key defined')
		
		result = []
		mac = hmac(self.secret_key, None, self.hash_method)
		for key, value in sorted (self.items(), key = lambda x: x[0]):
			result.append (quote_plus (key).encode ("utf8") + b"=" + self.quote(value))
			mac.update(b'|' + result[-1])			
		return (base64.b64encode(mac.digest()).strip() + b"?" + b'&'.join(result)).decode ("utf8")
	
	@classmethod
	def unserialize(cls, request, string, secret_key, setfunc, *args, **kargs):
		items = {}
		try:
			base64_hash, data = string.split(b'?', 1)
		except:
			base64_hash, data, items = b"", b"", None
		
		mac = hmac(secret_key, None, cls.hash_method)
		for item in data.split(b'&'):
			mac.update(b'|' + item)
			if not b'=' in item:
				items = None
				break
			key, value = item.split(b'=', 1)
			# try to make the key a string
			try:
				key = unquote_plus (key.decode ("utf8"))
			except UnicodeError:
				pass
			items[key] = value
		
		try:
			client_hash = base64.b64decode(base64_hash)
		except Exception:
			items = client_hash = None
		
		if items is not None and client_hash == mac.digest():
			try:
				for key, value in items.items():					
					items[key] = cls.unquote(value)
			except UnquoteError:
				items = {}
								
		else:
			items = {}
						
		return cls (request, items, secret_key, setfunc, False, *args, **kargs)		

class SecuredListValue (SecuredValue):
	KEY = "SECURELIST"	
	
	def set_default_data (self):
		self.dirty = True
		self.data = []
	
	def serialize(self):
		if self.secret_key is None:
			raise RuntimeError('no secret key defined')
								
		result = []
		mac = hmac(self.secret_key, None, self.hash_method)
		for value in sorted (self.data):
			result.append (self.quote (value))
			mac.update(b'|' + result[-1])
		return (base64.b64encode(mac.digest()).strip() + b"?" + b'&'.join(result)).decode ("utf8")
	
	@classmethod
	def unserialize(cls, request, string, secret_key, setfunc, *args, **kargs):
		items = []
		try:
			base64_hash, data = string.split(b'?', 1)
		except:
			base64_hash, data, items = b"", b"", None
		
		mac = hmac(secret_key, None, cls.hash_method)
		for item in data.split(b'&'):
			mac.update(b'|' + item)
			items.append (cls.unquote (item))
		
		try:
			client_hash = base64.b64decode(base64_hash)
		except Exception:
			items = client_hash = None
					
		return cls (request, items, secret_key, setfunc, False, *args, **kargs)
				
			
class Session (SecuredDictValue):
	default_session_timeout = 1200 # 20 min.	
	KEY = "SESSION"	
	
	def __init__ (self, request, data, secret_key, setfunc, new = True, session_timeout = 0):
		SecuredValue.__init__ (self, request, data, secret_key, setfunc, new)
		self.session_timeout = session_timeout and session_timeout or self.default_session_timeout
		self.__source_verified = False
		if self.data:
			self.validate ()
		
	def validate (self):	
		if not '_expires' in self.data:
			self.data = {}
			return
			
		if type (self.data ['_expires']) is tuple:
			expires, addr = self.data ['_expires']
			self.__source_verified = (addr == self.request.get_remote_addr ())				
		else:
			expires = self.data ['_expires']
			
		if time.time() > expires: # expired
			self.data = {}
			return
			
		if self.session_timeout + time.time () < expires: # too long, maybe fraud
			self.data = {}
			return
	
	def getv (self, k, v = None):
		if self.__source_verified:
			return self.get (k, v)
		return v	
		
	def source_verified (self):
		return self.__source_verified
				
	def recal_expires (self, expires):
		if expires is None:
			return self.session_timeout
		if expires == "now":
			return 0
		if expires == "never":
			raise ValueError("session must be specified expires seconds")
		return int (expires)
		
	def commit (self, expires = None, path = "/", domain = None):
		# always commit for extending/expiring expires
		expires = self.recal_expires (expires)
		self ["_expires"] = (time.time () + expires, self.request.get_remote_addr ())
		if len (self.data) == 1: # only have _expires, expire now
			expires = 0
		self.setfunc (self.KEY, self.serialize (), expires, path, domain)
		self.dirty = False
		
	def set_default_session_timeout (self, timeout):
		self.default_session_timeout = timeout
		
		
class MessageBox (SecuredListValue):
	KEY = "NOTIS"
	
	def __init__ (self, request, data, secret_key, setfunc, new = True):
		SecuredValue.__init__ (self, request, data, secret_key, setfunc, new)
		self.mid = -1
		
	def send (self, msg, category = "info", valid = 0, **extra):
		if self.data and self.mid == -1:
			self.mid = max ([n [0] for n in self.data])		
		self.mid += 1
		self.data.append ((self.mid, category, int (time.time ()), valid, msg, extra))
		self.dirty = True
	
	def remove (self, mid):
		index = 0
		found = False
		for n in self.data:
			if n [0] == mid:
				found = True
				break
			index += 1		
		if found:
			self.data.pop (index)
		self.dirty = True	
			
	def get (self, *wanted_category):
		now = int (time.time ())
		messages = []	
		not_expired = []
		
		for notice in self.data:
			how_old = now - notice [2]
			
			if notice [3] and how_old > notice [3]:
				# expired, drop
				continue
			
			if wanted_category and notice [1] not in wanted_category:
				not_expired.append (notice)
				continue
				
			if notice [3]:
				not_expired.append (notice)			
							
			messages.append (notice)			
		
		if len (self.data) != len (not_expired):
			self.data = not_expired
			self.dirty = True
			
		return messages
	
	def recal_expires (self, expires):						
		if self.data:
			return "never"
		return 0
	
	


from . import wsgi_handler, collectors
import re
import skitai
from skitai.protocol.http import request as http_request
from skitai.protocol.http import request_handler as http_request_handler
from skitai.protocol.http import response as http_response
from skitai.lib import compressors
from skitai.lib import producers
import time

post_max_size = wsgi_handler.Handler.post_max_size
upload_max_size = wsgi_handler.Handler.upload_max_size
PROXY_TUNNEL_KEEP_ALIVE = 120

class AsynTunnel:
	collector = None
	producer = None
	def __init__ (self, asyncon, handler):
		self.asyncon = asyncon
		self.handler = handler
		self.bytes = 0
		self.asyncon.set_terminator (None)
		
	def collect_incoming_data (self, data):
		#print (">>>>>>>>>>>>", data)
		self.bytes += len (data)		
		self.asyncon.push (data)
	
	def close (self):
		# will be closed by channel
		#print (">>>>>>>>>>>>> AsynTunnel xlosed", self.handler.request.uri)
		self.handler.channel_closed ()
		
	
class TunnelHandler:
	keep_alive = 120			
	def __init__ (self, asyncon, request, channel):		
		self.asyncon = asyncon		
		self.request = request
		self.channel = channel
		
		self.asyntunnel = AsynTunnel (asyncon, self)		
		self.channel.set_response_timeout	(PROXY_TUNNEL_KEEP_ALIVE)
		self.asyncon.set_network_delay_timeout (PROXY_TUNNEL_KEEP_ALIVE)
		self.channel.add_closing_partner (self.asyntunnel)
		
		self.bytes = 0
		self.stime = time.time ()
				
	def trace (self, name = None):
		if name is None:
			name = "tunnel://%s:%d" % self.asyncon.address
		self.channel.trace (name)
		
	def log (self, message, type = "info"):
		uri = "tunnel://%s:%d" % self.asyncon.address
		self.channel.log ("%s - %s" % (uri, message), type)
					
	def collect_incoming_data (self, data):	
		#print ("<<<<<<<<<<<<", data[:20])
		self.bytes += len (data)
		self.channel.push (data)
		
	def log_request (self):
		htime = (time.time () - self.stime) * 1000
		self.channel.server.log_request (
			'%s:%d CONNECT tunnel://%s:%d HTTP/1.1 200 %d/%d %dms %dms'
			% (self.channel.addr[0],
			self.channel.addr[1],			
			self.asyncon.address [0],
			self.asyncon.address [1],
			self.asyntunnel is not None and self.asyntunnel.bytes or 0,
			self.bytes,
			htime,
			htime
			)
		)
		
	def connection_closed (self, why, msg):
		#print ("------------Asyncon_disconnected", self.request.uri, self.bytes)
		# Disconnected by server mostly caused by Connection: close header				
		self.channel.close_when_done ()
		self.channel.current_request = None
		
	def channel_closed (self):
		#print ("------------channel_closed", self.request.uri, self.bytes)
		self.asyncon.handler = None
		self.asyncon.disconnect ()
		self.asyncon.end_tran ()


class ProxyRequestHandler (http_request_handler.RequestHandler):
	def __init__ (self, asyncon, request, callback, client_request, collector, connection = "keep-alive"):
		http_request_handler.RequestHandler.__init__ (self, asyncon, request, callback, "1.1", connection)
		self.collector = collector
		self.client_request = client_request
		self.is_tunnel = False
		self.new_handler = None
			
	def add_reply_headers (self):		
		for line in self.response.get_headers ():			
			try: k, v = line.split (": ", 1)
			except:	continue			
			ll = k.lower ()
			if ll in ("expires", "date", "connection", "keep-alive", "content-length", "transfer-encoding", "content-encoding", "age", "vary"):
				continue
			if ll == "server":
				self.client_request.response.update ("Server", v.strip () + ", " + skitai.NAME)
				continue
			self.client_request.response [k] = v.strip ()			
	
	ESTABLISHED = b"HTTP/1.1 200 Connection Established\r\nServer: " + skitai.NAME.encode ("utf8")
	def has_been_connected (self):
		if self.request.method == "connect":
			self.buffer =	self.ESTABLISHED
			self.found_terminator ()
	
	def will_open_tunneling (self):
		return self.response.code == 200 and self.response.msg.lower () == "connection established"
	
	def connection_closed (self, why, msg):
		if self.client_request.channel is None:
			return self.close_case_with_end_tran ()			
		return http_request_handler.RequestHandler.connection_closed (self, why, msg)
					
	def close_case (self):
		# unbind readable/writable methods
		if self.asyncon:
			self.asyncon.ready = None
			self.asyncon.affluent = None			
			if self.client_request.channel:
				self.client_request.channel.ready = None
				self.client_request.channel.affluent = None			
			
			self.asyncon.handler = self.new_handler			
		
		if self.callback:			
			self.callback (self)
	
	def found_end_of_body (self):
		# proxy should not handle authorization		
		if self.will_be_close ():
			self.asyncon.disconnect ()
		self.close_case_with_end_tran ()
	
	def create_tunnel (self):
		self.asyncon.established = True		
		self.new_handler = TunnelHandler (self.asyncon, self.request, self.client_request.channel)
		# next_request = (new request, new terminator)
		self.client_request.response.done (globbing = False, compress = False, next_request = (self.new_handler.asyntunnel, None)) 
			
	def create_response (self):		
		if not self.client_request.channel: return

		buffer, self.buffer = self.buffer, b""
		accept_gzip = ""
		accept_encoding = self.client_request.get_header ("Accept-Encoding")
		if accept_encoding and accept_encoding.find ("gzip") != -1:
			accept_gzip = "gzip"
		
		try:
			self.response = ProxyResponse (self.request, buffer.decode ("utf8"), accept_gzip, self.client_request, self.asyncon)
		except:
			#print (buffer)
			self.log ("response header error: `%s`" % repr (buffer [:80]), "error")
			self.asyncon.handle_close (708, "Response Header Error")
			return

		if self.is_continue_response ():
			# maybe response code is 100 continue
			return
		
		if self.will_open_tunneling ():
			self.create_tunnel ()
			self.close_case ()
			return
				
		self.client_request.response.start (self.response.code, self.response.msg)
		self.add_reply_headers ()
		
		if self.response.is_gzip_compressed ():
			self.client_request.response ["Content-Encoding"] = "gzip"
		
		if self.response.body_expected ():
			self.client_request.response.push (self.response)
			#self.client_request.channel.add_closable_producer (self.response)
			# in relay mode, possibly delayed
			self.client_request.channel.ready = self.response.ready			
			self.asyncon.affluent = self.response.affluent
		
		self.client_request.response.done (globbing = False, compress = False)
	
	def start (self):
		if not self.client_request.channel: return
		
		self.buffer, self.response = b"", None
		self.asyncon.set_terminator (b"\r\n\r\n")	
		
		if self.request.method != "connect":
			for buf in self.get_request_buffer ():
				self.asyncon.push (buf)
			if self.collector:
				self.collector.reuse_cache ()
				if not self.collector.got_all_data:
					self.asyncon.ready = self.collector.ready
					self.client_request.channel.affluent = self.collector.affluent
					# don't init_send cause of producer has no data yet
				self.asyncon.push_with_producer (self.collector, init_send = False)
				
		self.asyncon.begin_tran (self)
								
	def get_request_buffer (self):
		hc = {}
		address = self.request.get_address ()
		if address [1] in (80, 443):
			hc ["Host"] = "%s" % address [0]
		else:
			hc ["Host"] = "%s:%d" % address
			
		hc ["Connection"] = self.connection
		hc ["Accept-Encoding"] = "gzip"
		
		method = self.request.get_method ()			
		additional_headers = self.client_request.get_headers ()
		
		if additional_headers:
			for line in additional_headers:
				k, v = line.split (": ", 1)
				ll = k.lower ()
				if ll in ("connection", "keep-alive", "accept-encoding", "host"):
					continue
				hc [k] = v
				
		hc ["X-Forwarded-For"] = "%s" % self.client_request.get_remote_addr ()
		hc ["X-Proxy-Agent"] = skitai.NAME
				
		req = "%s %s HTTP/%s\r\n%s\r\n\r\n" % (
			method,
			self.request.path,
			self.http_version,
			"\r\n".join (["%s: %s" % x for x in list(hc.items ())])			
		)
		
		#print ("####### SKITAI => SERVER ##########################")
		#print (req)
		#print ("---------------------------------")
		return [req.encode ("utf8")]

	
class ProxyResponse (http_response.Response):
	SIZE_LIMIT = 2**24
	
	def __init__ (self, request, header, accept_gzip, client_request, asyncon):		
		self.client_request = client_request
		self.asyncon = asyncon
		self.accept_gzip = accept_gzip		
		self.request = request
		self.header_s = header		
		if header [:2] == "\r\n":
			header = header [2:]
		header = header.split ("\r\n")		
		self.response = header [0]
		self.header = header [1:]
		self._header_cache = {}
		self.flushed_time = 0
		self.client_request.producer = self
		self.version, self.code, self.msg = http_response.crack_response (self.response)
		self.p, self.u = None, None
		self.decompressor = None
		self.gzip_compressed = False	
		self.make_decompressor = False
		
		content_encoding = self.get_header ("Content-Encoding")			
		if content_encoding == "gzip":
			if self.accept_gzip:
				self.gzip_compressed = True
			else:	
				self.make_decompressor = True
						
		self.size = 0
		self.got_all_data = False
		
		self.reqtype = "HTTP"				
		if self.client_request.get_header ("cookie"):
			self.max_age = 0
		else:	
			self.set_max_age ()
	
	def body_expected (self):
		cl = self.get_header ("Content-Length")
		if cl == 0:
			self.got_all_data = True
			return False
		
		te = self.get_header ("Transfer-Encoding")
		if cl is None and te != "chunked":
			hv = self.version
			cn = self.get_header ("Connection")			
			if cn is None:
				if hv == "1.0": cn = "close"
				else: cn = "keep-alive"	
			else:
				cn = cn.lower ()			
			if cn == "keep-alive":				
				self.got_all_data = True
				return False
		
		if self.p is None:
			self.p, self.u = http_response.getfakeparser (cache = self.max_age)
			if self.make_decompressor:
				self.decompressor = compressors.GZipDecompressor ()
			
		return True
		
	def init_buffer (self):
		# do this job will be executed in body_expected ()
		pass
		
	def is_gzip_compressed (self):
		return self.gzip_compressed
	
	def close (self):
		self.client_request.producer = None
		try: self.u.data = []
		except AttributeError: pass		
		#self.asyncon.disconnect ()		
		#self.asyncon.end_tran ()
		self.asyncon.handle_close (710, "Channel Closed")
			
	def affluent (self):
		# if channel doesn't consume data, delay recv data
		return len (self.u.data) < 1000
		
	def ready (self):
		# if exist consumable data or wait		
		return len (self.u.data) or self.got_all_data
		
	def more (self):
		self.flushed_time = time.time ()
		return self.u.read ()


class Collector (collectors.FormCollector):
	# same as asyncon ac_in_buffer_size
	ac_in_buffer_size = 4096
	asyncon = None
	def __init__ (self, handler, request):
		self.handler = handler
		self.request = request
		self.data = []
		self.cache = []
		self.cached = False
		self.got_all_data = False
		self.length = 0 
		self.content_length = self.get_content_length ()
	
	def reuse_cache (self):
		self.data = self.cache + self.data		
		self.cache = []
			
	def start_collect (self):	
		if self.content_length == 0:
			return self.found_terminator ()
			
		if self.content_length <= post_max_size: #5M
			self.cached = True
		
		self.request.channel.set_terminator (self.content_length)
	
	def close (self):
		# channel disconnected
		self.data = []
		self.cache = []
		self.request.collector = None
		
		# abort immediatly
		if self.asyncon:
			self.asyncon.handle_close (710, "Channel Closed")
				
	def collect_incoming_data (self, data):
		#print "proxy_handler.collector << %d" % len (data), id (self)
		self.length += len (data)
		self.data.append (data)

	def found_terminator (self):
		self.request.channel.set_terminator (b'\r\n\r\n')
		self.got_all_data = True
		self.request.collector = None
		# don't request.collector = None => do it at callback ()
		# because this collector will be used in Request.continue_start() later
	
	def get_cache (self):
		return b"".join (self.cache)
	
	def affluent (self):
		# if channel doesn't consume data, delay recv data		
		return len (self.data) < 1000
		
	def ready (self):
		return len (self.data) or self.got_all_data
	
	def more (self):
		if not self.data:
			return b""
						
		data = []
		tl = 0
		while self.data:
			tl += len (self.data [0])
			if tl > self.ac_in_buffer_size:
				break
			data.append (self.data.pop (0))
		
		if self.cached:			
			self.cache += data
		#print "proxy_handler.collector.more >> %d" % tl, id (self)
		return b"".join (data)
		
			
class Handler (wsgi_handler.Handler):
	def __init__ (self, wasc, clusters, cachefs = None):
		self.wasc = wasc
		self.clusters = clusters
		self.cachefs = cachefs
		#self.cachefs = None # DELETE IT!		
		self.q = {}
				
	def match (self, request):
		uri = request.uri.lower ()
		if uri.startswith ("http://") or uri.startswith ("https://") or uri.startswith ("ws://") or uri.startswith ("wss://"):
			return 1
		if request.command == "connect":
			return 1
		return 0
		
	def handle_request (self, request):
		collector = None
		if request.command in ('post', 'put'):
			ct = request.get_header ("content-type")
			if not ct: ct = ""
			current_post_max_size = ct.startswith ("multipart/form-data") and upload_max_size or post_max_size
			collector = self.make_collector (Collector, request, current_post_max_size)
			if collector:
				request.collector = collector
				collector.start_collect ()
			else:
				return # error was already called in make_collector ()							
		self.continue_request(request, collector)
					
	def continue_request (self, request, collector, asyncon = None):		
		if self.is_cached (request, collector is not None):
			return
		
		try:
			if asyncon is None:		
				if request.command == "connect":
					asyncon_key = "tunnel://" + request.uri + "/"
				else:
					asyncon_key = request.uri					
				asyncon = self.clusters ["__socketpool__"].get (asyncon_key)
			
			req = http_request.HTTPRequest (request.uri, request.command, collector is not None, logger = self.wasc.logger.get ("server"))				
			r = ProxyRequestHandler (asyncon, req, self.callback, request, collector)			
			if collector:
				collector.asyncon = asyncon
			r.start ()
						
		except:
			self.wasc.logger.trace ("server")			
			request.response.error (500, "", "Proxy request has been failed.")
		
	def is_cached (self, request, has_data):		
		if has_data:
			return False
		if request.get_header ("cookie") is not None:
			return False
		if request.get_header ("progma") == "no-cache" or request.get_header ("cache-control") == "no-cache":
			request.response ["X-Cache-Lookup"] = "PASSED"
			return False
		
		if self.cachefs:
			try:
				accept_encoding = request.get_header ("accept-encoding")
				hit, compressed, max_age, content_type, content = self.cachefs.get (request.uri, "", accept_encoding and accept_encoding.find ("gzip") != -1)
						
			except:
				self.wasc.logger.trace ("server")	
				return False
			
			else:
				if hit:
					if hit == 1:
						request.response ["X-Cache-Lookup"] = "MEM_HIT"
					else:
						request.response ["X-Cache-Lookup"] = "HIT"
					if content_type:
						request.response ["Content-Type"] = content_type					
					request.response ["Cache-Control"] = "max-age=%d" % max_age
					if compressed:
						request.response ["Content-Encoding"] = "gzip"
					request.response.push (content)
					request.response.done ()
					return True
					
		request.response ["X-Cache-Lookup"] = "MISS"
		return False
	
	def save_cache (self, request, handler):
		try:			
			if self.cachefs and handler.response.max_age:
				self.cachefs.save (
					request.uri, handler.request.data, 
					handler.response.get_header ("content-type"), handler.response.get_content (), 
					handler.response.max_age, request.response ["Content-Encoding"] == "gzip"
				)
				
		except:
			self.wasc.logger.trace ("server")
	
	def dealloc (self, request, handler):
		handler.callback = None				
		handler.response = None
		handler.collector = None
		request.collector = None
		request.producer = None
		request.response = None # break back ref.		
		del handler
			
	def callback (self, handler):
		response, request = handler.response, handler.client_request
		
		if request.channel:
			if response.code >= 700:
				request.response.error (506, response.msg)
			
			else:
				try:
					self.save_cache (request, handler)
				except:
					self.wasc.logger.trace ("server")
		
		self.dealloc (request, handler)
		


#-------------------------------------------------------
# Basic Blade Server Architecture R2
# Hans Roh (hansroh@gmail.com)
# 2014.4.24 Python 2.7 port (from 2.4)
#-------------------------------------------------------

HTTPS = True
from skitai.client import adns
import sys, time, os, threading
from . import http_server, wsgi_apps, rcache
from skitai import lifetime
from warnings import warn
from . import https_server
from skitai import start_was
if os.name == "nt":	
	from . import schedule			
from .handlers import proxy_handler, pingpong_handler, vhost_handler
from .threads import threadlib, trigger
from skitai.lib import logger, confparse, pathtool, flock
from .rpc import cluster_dist_call, cachefs		
from skitai.client import socketpool
import socket
import signal
import multiprocessing
from . import wsgiappservice
from .dbi import cluster_dist_call as dcluster_dist_call
from skitai.dbapi import dbpool


class Loader:
	def __init__ (self, config, logpath, varpath, debug = 0):
		self.config = config
		self.instance_name = os.path.split (config)[-1][:-5]
		self.logpath = logpath
		self.varpath = varpath
		self.debug = debug
		self.num_worker = 1
		self.wasc = wsgiappservice.WAS
		self.ssl = False
		self.ctx = None
		self._exit_code = None
		self.config_logger (self.logpath)
		self.WAS_initialize ()
		self.configure ()
		self.WAS_finalize ()
	
	def configure (self):
		raise SystemExit("configure must be overided")
	
	def set_num_worker (self, num):	
		if os.name == "nt":
			num = 1
		self.num_worker = num
			
	def WAS_initialize (self):		
		self.wasc.log_base_path = os.path.split (os.path.split (self.logpath)[0])[0]
		self.wasc.var_base_path = os.path.split (os.path.split (self.varpath)[0])[0]
		self.wasc.register ("debug", self.debug)
		self.wasc.register ("plock", multiprocessing.RLock ())
		self.wasc.register ("clusters",  {})
		self.wasc.register ("clusters_for_distcall",  {})
		
	def WAS_finalize (self):
		global the_was
		
		self.wasc.register ("lock", threading.RLock ())
		self.wasc.register ("lifetime", lifetime)		
		
		adns.init (self.wasc.logger.get ("server"))		
		socketfarm = socketpool.SocketPool (self.wasc.logger.get ("server"))
		self.wasc.clusters ["__socketpool__"] = socketfarm
		self.wasc.clusters_for_distcall ["__socketpool__"] = cluster_dist_call.ClusterDistCallCreator (socketfarm, self.wasc.logger.get ("server"))		
		
		dp = dbpool.DBPool (self.wasc.logger.get ("server"))
		self.wasc.clusters ["__dbpool__"] = dp
		self.wasc.clusters_for_distcall ["__dbpool__"] = dcluster_dist_call.ClusterDistCallCreator (dp, self.wasc.logger.get ("server"))
		
		if not hasattr (self.wasc, "threads"):
			for attr in ("map", "rpc", "rest", "wget", "lb", "db", "dlb", "dmap"):
				delattr (self.wasc, attr)
		
		start_was (self.wasc)
		
	def config_cachefs (self, cache_dir): 
		if cache_dir:
			self.wasc.register ("cachefs",  cachefs.CacheFileSystem (cache_dir))
	
	def config_rcache (self, maxobj = 2000):
		rcache.start_rcache (maxobj)
		self.wasc.register ("rcache", rcache.the_rcache)		
		
	def config_certification (self, certfile, keyfile = None, pass_phrase = None):
		if not HTTPS:
			return
		if not os.path.isfile (certfile):
			_certpath = os.path.join (os.path.split (os.path.split (self.config)[0])[0], "cert")
			certfile = os.path.join (_certpath, certfile)
			if keyfile:
				keyfile = os.path.join (_certpath, keyfile)
		self.ctx = https_server.init_context (certfile, keyfile, pass_phrase)		
		self.ssl = True
				
	def config_webserver (self, port, ip = "", name = "", ssl = False, keep_alive = 10, response_timeout = 10):
		# maybe be configured	at first.
		if ssl and not HTTPS:
			raise SystemError("Can't start SSL Web Server")
		
		if not name:
			name = self.instance_name		
		http_server.configure (name, response_timeout, keep_alive)
				
		if ssl and self.ctx is None:
			raise ValueError("SSL ctx not setup")
		
		if ssl:
			server_class = https_server.https_server			
		else:	
			server_class = http_server.http_server
		
		if self.ssl:
			httpserver = server_class (ip and ip or "", port, self.ctx, self.wasc.logger.get ("server"), self.wasc.logger.get ("request"))	
		else:
			httpserver = server_class (ip and ip or "", port, self.wasc.logger.get ("server"), self.wasc.logger.get ("request"))	
		
		self.wasc.register ("httpserver", httpserver)
		
		#fork here 
		_exit_code = self.wasc.httpserver.fork_and_serve (self.num_worker)				
		if _exit_code is not None:
			self.handle_exit_code (_exit_code)
		
	def handle_exit_code (self, _exit_code):		
		self._exit_code = _exit_code				
		try:
			os.wait ()
		except OSError:	
			pass			
	
	def get_exit_code (self):	
		return self._exit_code
					
	def config_scheduler (self, conffile):
		if os.name == "nt":
			scheduler = schedule.Scheduler (self.wasc, conffile, self.wasc.logger.get ("server"))		
			self.wasc.register ("scheduler", scheduler)
	
	def config_logger (self, path):
		if path is not None:
			media = ["file"]
		else:
			media = ["screen"]	
		self.wasc.register ("logger", wsgiappservice.Logger (media, path))
		
		if os.name != "nt":
			def hUSR1 (signum, frame):	
				self.wasc.logger.rotate ()
			signal.signal(signal.SIGUSR1, hUSR1)
	
	def config_threads (self, numthreads = 0):
		if numthreads > 0:
			trigger.start_trigger ()
			queue = threadlib.request_queue2 ()
			tpool = threadlib.thread_pool (queue, numthreads, self.wasc.logger.get ("server"))
			self.wasc.register ("queue",  queue)
			self.wasc.register ("threads", tpool)
					
	def add_cluster (self, clustertype, clustername, clusterlist, ssl = 0):
		if ssl in ("1", "yes"): ssl = 1
		else: ssl = 0
		self.wasc.add_cluster (clustertype, clustername, clusterlist, ssl = ssl)
			
	def install_handler (self, routes = {}, proxy = False, static_max_age = 300, upload_max_size = 0):		
		self.wasc.add_handler (1, pingpong_handler.Handler)		
		if proxy:
			self.wasc.add_handler (1, proxy_handler.Handler, self.wasc.clusters, self.wasc.cachefs)		
		vh = self.wasc.add_handler (1, vhost_handler.Handler, self.wasc.clusters, self.wasc.cachefs, static_max_age, upload_max_size)		
		
		current_rule = "default"
		for line in routes:
			if line.startswith (":"):
				current_rule = line [1:]
			elif line.startswith (";") or line.startswith ("#"):
				continue
			else:
				vh.add_route (current_rule, line)	
			
	def run (self):
		if self._exit_code is not None: 
			return self._exit_code # master process
			
		try:
			try:
				lifetime.loop ()
			except:
				self.wasc.logger.trace ("server")					
		finally:
			self.close ()
		
		return None # worker process
		
	def close (self):
		for attr, obj in list(self.wasc.objects.items ()):
			if attr == "logger": continue
			try:
				self.wasc.logger ("server", "[info] clenaup %s" % attr)
				obj.cleanup ()
				del obj
			except AttributeError:
				pass
			except:
				self.wasc.logger.trace ("server")
		
		if self.wasc.httpserver.worker_ident == "master":
			self.wasc.logger ("server", "[info] cleanup done, closing logger... bye")
			try:
				self.wasc.logger.close ()
				del self.wasc.logger
			except:
				pass




import base64
from hashlib import md5
import re
try:
	from urllib.parse import urlparse, quote_plus	
except ImportError:
	from urlparse import urlparse
	from urllib import quote_plus	
import time
from skitai.protocol.http import util
from . import localstorage


class EURL:
	keywords = ('from',)
	methods = ("get", "post", "head", "put", "delete", "options", "trace", "connect", "upload")
	dft_port_map = {'http': 80, 'https': 443, 'ftp': 21}
	DEFAULT_USER_AGENT = "Mozilla/5.0 (compatible; Skitaibot/0.1a)"	
	
	def __init__ (self, surl, data = {}):
		self.surl = surl
		self.data = data
		self.user = {}
		self.http = {"http-version": "1.1", "http-connection": "close"}
		self.info = {"depth": 0, "moved": 0, "retrys": 0}
		self.header = {"Cache-Control": "max-age=0"}
		self.parse ()
	
	def __str__ (self):
		return self ["rfc"]
		
	def set_header_from_option (self, k, v):
		"Possibles: http-[auth|form|version|connection|cookie|content-type|user-agent|proxy|tunnel]"
		if k [:5] == "head-":
			self.set_header (k [5:], v)		
		elif k ==	"http-cookie":
			self.set_header ("Cookie", v)
		elif k ==	"http-content-type":
			self.set_header ("Content-Type", v)
		elif k ==	"http-referer":
			self.set_header ("Referer", v)
		self.http [k] = v		
				
	def set_header (self, k, v):
		self.header [k] = v
	
	def del_header (self, key):
		for k, v in list (self.header.items ()):
			if key.lower () == k.lower ():
				del self.header [k]
				
	def get_header (self, key = None, default = None):
		if key:
			for k, v in list (self.header.items ()):
				if key.lower () == k.lower ():
					return v
			return default
		return self.header
						
	def __setitem__ (self, k, v):
		if k [:5] in ("http-", "head-", "wsoc-"):
			self.set_header_from_option (k, v)
		else:	
			self.info [k] = v
	
	def __delitem__ (self, k):
		try:
			del self.info [k]
		except KeyError:
			del self.http [k]	
		
	def __getitem__ (self, k):
		try:
			return self.info [k]
		except KeyError:
			try:
				return self.http [k]
			except KeyError:	
				return None
	
	def get (self, k, d):
		return self.info.get (k) or self.http.get (k, d)
		
	def items (self):
		return list(self.info.items ()) + list(self.http.items ())
	
	def has_key (self, k):
		return k in self.info or k in self.http
			
	def __str__ (self):
		return self ["url"]
	
	def to_version_11 (self):
		self ["http-version"] = "1.1"
		if self ["http-connection"].lower () == "close":
			del self ["http-connection"]
	
	def inc_retrys (self):
		self.info ["retrys"] = self ["retrys"] + 1
	
	def dec_retrys (self):
		self.info ["retrys"] = self ["retrys"] - 1
				
	def inherit (self, surl, moved = False):
		eurl = EURL (surl)
				
		# inherit user-data
		for k, v in list(self.user.items ()):
			if k not in eurl.user: # prevent overwriting
				eurl.user [k] = v
		
		# inherit info
		eurl ["http-auth"] = self ["http-auth"]
		eurl ["referer"] = eurl ["http-referer"] = self ["rfc"]
		eurl ["referer_id"] = self ["page_id"]
		
		if moved:
			eurl ["moved"] = self ["moved"] + 1		
		else:
			eurl ["depth"] = self ["depth"] + 1
			
		# inherit http, header
		for k, v in list(self.http.items ()):
			if k [:5] == "head-":
				eurl [k] = v
			else:
				if k [5:] in ("cookie", "referer", "content-type", "form"):
					continue
				if k not in eurl.http:
					eurl [k] = v										
		return eurl
		
	def parse (self):
		d = {}
		current_key = None
		for token in self.surl.split(" "):
			tk = token.lower()
			if tk in self.keywords or tk in self.methods:
				current_key = tk
				d[current_key] = ''
			
			elif tk [:7] in ("--with-", "--head-", "--http-", "--wsoc-"):				
				current_key = token
				d[current_key] = ''
				
			else:
				try: 
					if d[current_key]:
						 d[current_key] += ' ' + token
					else:
							d[current_key] = token
							
				except KeyError:
					current_key = 'get'					
					d[current_key] = token
		
		if self.data:
			d ["http-form"] =  util.dictencode (self.data)
			self.data = {}
						
		self._set (d)

	def	_set (self, d):
		for k, v in list(d.items ()):
			if k [:7] == "--with-":
				k = k [7:]
				self.user [k] = v.strip ()
			elif k in self.methods:
				self ['method'] = k
				self ['url'] = v.strip ()
			elif k == "from":
				self ["referer"] = v.strip ()
				self ["http-referer"] = v.strip ()				
			elif k [:7] in ("--http-", "--head-", "--wsoc-"):
				self [k [2:]] = v.strip ()
			
		self ['scheme'], self ['netloc'], self ['script'], self ['params'], self ['querystring'], self ['fragment'] = urlparse (self ['url'])
		if self ["http-form"] and self ['method'] not in ("post", "put"):
			raise ValueError("Form exists but method isn't post or get")
		if self ["method"] in ("post", "put") and not self ["http-form"]:
			raise ValueError("No form data")
		if not self ["http-form"] and self ["http-content-type"]:
			raise ValueError("Needn't content-type")
		if self ["http-form"] and self ["method"] == "post" and self.get_header ("content-type") is None:
			self ["http-content-type"] = "application/x-www-form-urlencoded; charset=utf-8"		
			
		if not self ['script']: 
			self ['script'] = '/'
		if self ['querystring']:
			try: self ['querystring'] = uitl.strencode (self ['querystring'])
			except: pass
		
		uri = self ['script'] #real request uri
		if self ['params']:
			uri += ';' + self ['params']
		if self ['querystring']:
			uri += '?' + self ['querystring']
		self ['uri'] = uri
		
		position = self ['netloc'].find('@')
		if position > -1:
			self ["http-auth"], self ['netloc'] = self ['netloc'].split ("@", 1)
		
		if self ["http-auth"]:
			try: 
				self ['username'], self ['password'] = self ["http-auth"].split (':', 1)
			except ValueError:	
				pass
			else:	
				self ["http-auth"] = (self ['username'], self ['password'])
				
		try:
			self ['netloc'], self ['port'] = self ['netloc'].split (':', 1)
		except ValueError:
			self ['netloc'], self ['port'] = self ['netloc'], self.dft_port_map [self ['scheme']]
		
		self ['netloc'] = self ['netloc'].lower ()
		try: self ['port'] = int (self ['port'])
		except: self ['port'] = 80
		
		netloc = self ['netloc']
		netloc2 = netloc.split (".")
		
		if len (netloc2 [-1]) == 3:
			self ['domain'] = ".".join (netloc2 [-2:])
			self ['host'] = ".".join (netloc2 [:-2])
			if not self ['host']:
				self ['host'] = "www"
		elif len (netloc2 [-1]) == 2:
			if len (netloc2) == 2:
				self ['domain'] = ".".join (netloc2)
				self ['host'] = "www"
			elif len (netloc2) == 3:
				if netloc2 [0] == "www":
					self ['domain'] = ".".join (netloc2 [-2:])
				else:
					self ['domain'] = ".".join (netloc2)
				self ['host'] = "www"
			else:
				self ['domain'] = ".".join (netloc2 [-3:])
				self ['host'] = ".".join (netloc2 [:-3])
			
		try: dft_port = self.dft_port_map [self ['scheme']]
		except KeyError:
			self ['rfc'] = '%s://%s:%d%s' % (self ['scheme'], self ['netloc'], self ['port'], self ['uri'])
		else:
			if dft_port == self ['port']:
				self ['rfc'] = '%s://%s%s' % (self ['scheme'], self ['netloc'], self ['uri'])
			else:
				self ['rfc'] = '%s://%s:%d%s' % (self ['scheme'], self ['netloc'], self ['port'], self ['uri'])
		
		self ["page_id"] = self.geneate_page_id ()
		
		ua = self ["http-user-agent"]
		if ua is None:			
			self ["http-user-agent"] = self.DEFAULT_USER_AGENT
		
		# set cookie
		ls = localstorage.localstorage
		cookie = self ["http-cookie"]
		if cookie:
			ls.set_cookie_from_data (self ["rfc"], cookie)					
		cookie = ls.get_cookie_as_string (self ["rfc"])		
		if cookie:
			self ["http-cookie"] = cookie
		
	def get_connection (self):
		return self ["http-connection"]
				
	def get_useragent (self):
		return self ["http-user-agent"]
	
	def __sort_args (self, data):
		args = {}
		for arg in data.split ("&"):
			try: k, v = arg.split ("=")
			except: continue
			if v:
				args [k] = v
						
		args = list(args.items ())
		args.sort ()
		
		argslist = []
		for k, v in args:
			argslist.append ("%s=%s" % (k, v))
		
		argslist.sort ()		
		return "&".join (argslist)

	def geneate_page_id (self):		
		signature = self ['method'] + "|" + self ['netloc'] + "|" + str (self ["port"]) + "|" + self ['script'] + "|" + self ['params']
		if self ['querystring']:
			signature += "|" + self.__sort_args (self ['querystring'])
		if self ['http-form']:
			signature += "|" + self.__sort_args (self ['http-form'])
		return md5 (signature.encode ("utf8")).hexdigest ()
	
	def show (self):
		for k, v in list(self.info.items()):
			print("-%s: %s" % (k.upper(), v))




def test_crack_uql():
	d = (
		  "get http://hans:whddlgkr@target_site.com:8089/rpc2/agent.status?a=x#eter "
		  "with-ua ELAB robotec. with-sid 34543"
		  )
	EURL (d).show()
	print("------------")
	d = "post http://www.dsfds.com/asdsa.aps?asdas=1&sdfds=1 with-form a=a&b=432432"
	EURL (d).show()
	

if __name__ == "__main__":
	test_crack_uql ()



"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""

import skitai
__VER__ = skitai.VERSION

import sys
import os
import shutil, glob
from warnings import warn
try:
	from setuptools import setup
except ImportError:	
	from distutils.core import setup

if sys.argv[-1] == 'publish':
	if os.name == "nt":
		os.system('python setup.py sdist upload') # bdist_wininst --target-version=2.7
	else:		
		os.system('python setup.py sdist upload')
	sys.exit()

classifiers = [
  'License :: OSI Approved :: BSD License',
  'Development Status :: 4 - Beta',
  'Topic :: Internet :: WWW/HTTP',
	'Topic :: Internet :: WWW/HTTP :: HTTP Servers',				
	'Environment :: Console',
	'Environment :: No Input/Output (Daemon)',
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',
	'Intended Audience :: Science/Research',
	'Programming Language :: Python',
	'Programming Language :: Python :: 2.7',
	'Programming Language :: Python :: 3'
]

if "install" in sys.argv or "develop" in sys.argv:
	PY_MAJOR_VERSION = sys.version_info [0]
	if PY_MAJOR_VERSION == 3:
		if os.path.isfile ("skitai/lib/py2utils.py"):
			os.remove ("skitai/lib/py2utils.py")
			try: os.remove ("skitai/lib/py2utils.pyc")
			except OSError: pass
	else:
		if not os.path.isfile ("skitai/lib/py2utils.py"):
			with open ("skitai/lib/py2utils.py", "w") as f:
				f.write ("def reraise(type, value, tb):\n\traise type, value, tb\n")			
		
packages = [
	'skitai',
	'skitai.client',
	'skitai.dbapi',
	'skitai.lib',	
	'skitai.protocol',
	'skitai.protocol.dns',
	'skitai.protocol.dns.pydns',
	'skitai.protocol.http',	
	'skitai.protocol.ws',	
	'skitai.requests',
	'skitai.protocol.smtp',
	'skitai.server',
	'skitai.server.dbi',
	'skitai.server.handlers',
	'skitai.server.rpc',
	'skitai.server.threads',
	'skitai.saddle'
]

package_dir = {
	'skitai': 'skitai',
	'skitai.server': 'skitai/server'
}

skitaid_files = [
	"README.md",
	"skitaid/bin/*.py",
	"skitaid/pub/sample/*.py",	
	"skitaid/pub/sample/static/*.*",	
	"skitaid/pub/sample/templates/*.*",	
	"skitaid/etc/init/skitaid.conf",
	"skitaid/etc/init.d/skitaid", 
	"skitaid/etc/skitaid/skitaid.conf",
	"skitaid/etc/skitaid/servers-available/README.txt", 
	"skitaid/etc/skitaid/servers-enabled/sample.conf",
	"skitaid/etc/skitaid/cert/*.*",
	"protocol/dns/*.txt",
	"protocol/dns/pydns/*.txt",
	"requests/*.txt",
	"tools/benchmark/*.py",
	"tools/benchmark/*.txt",
	"tools/benchmark/*.ini",
]

package_data = {
	"skitai": skitaid_files
}

with open ("README.txt") as f:
	ldesc = f.read ()

setup(
	name='skitai',
	version=__VER__,
	description='Skitai WSGI App Engine',
	long_description = ldesc,
	author='Hans Roh',
	author_email='hansroh@gmail.com',
	url='https://gitlab.com/hansroh/skitai/wikis/home',
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	license='BSD',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/skitai",
	install_requires = ["jinja2"],
	classifiers=classifiers
)

def mkdir (tdir, mod = -1):
	if os.path.isdir (tdir): return
	chain = [tdir]
	while 1:
		tdir, last = os.path.split (tdir)			
		if not last: break
		chain.insert (0, tdir)
	
	for dir in chain [1:]:
		try: 
			os.mkdir (dir)
			if os.name == "posix" and mod != -1:
				os.chmod (dir, mod)				
		except OSError as why:
			if why.errno in (17, 183): continue
			else: raise

if "install" in sys.argv or "develop" in sys.argv:
	if os.name == "nt":
		mkdir ("c:\\skitaid")
		mkdir ("c:\\skitaid\\var")
		mkdir ("c:\\skitaid\\bin")
		mkdir ("c:\\skitaid\\log")
		mkdir ("c:\\skitaid\\pub\\sample\\static")
		mkdir ("c:\\skitaid\\pub\\sample\\templates")
		mkdir ("c:\\skitaid\\etc\\cert")
		mkdir ("c:\\skitaid\\etc\\servers-available")
		mkdir ("c:\\skitaid\\etc\\servers-enabled")		
		
		conf = "c:\\skitaid\\etc\\skitaid.conf"
		if not os.path.isfile (conf):
			shutil.copyfile ("skitai\\skitaid\\etc\\skitaid\\skitaid.conf", conf)
		
		pathes = ["cert\\README.txt", "servers-available\\README.txt", "servers-enabled\\sample.conf"]		
		for path in pathes:
			target = os.path.join ("c:\\skitaid\\etc", path)
			try: 
				os.remove (target)
			except WindowsError as why:
				if why.errno == 2: pass
			shutil.copyfile (os.path.join ("skitai\\skitaid\\etc\\skitaid", path), target)
		
		for fn in os.listdir ("skitai\\skitaid\\bin"):
			target = os.path.join ("c:\\skitaid\\bin", fn)
			try: os.remove (target)
			except WindowsError as why:
				if why.errno == 2: pass	
			shutil.copyfile (os.path.join ("skitai\\skitaid\\bin", fn), target)
		
		for fn in ("static\\index.html",):
			target = os.path.join ("c:\\skitaid\\pub\\sample", fn)
			try: os.remove (target)			
			except WindowsError as why:
				if why.errno == 2: pass	
					
		for fn in ("webapp.py", "static\\reindeer.jpg", "templates\\documentation.html", "templates\\index.html", "templates\\websocket.html"):
			target = os.path.join ("c:\\skitaid\\pub\\sample", fn)
			try: os.remove (target)
			except WindowsError as why:
				if why.errno == 2: pass	
			shutil.copyfile (os.path.join ("skitai\\skitaid\\pub\\sample", fn), target)
			
		print("\n\n======================================")
		print("Installation Complete")
		print("--------------------------------------")	
		print("Please run below command in your commands prompt with administator privilege\n")
		print("  cd /d c:\\skitaid\\bin")
		print("  python install-win32-service.py --startup auto install")
		print("  python install-win32-service.py start")
		print("  then check http://localhost:5000\n\n")
	
	else:
		mkdir ("/etc/skitaid")
		mkdir ("/etc/skitaid/cert")
		mkdir ("/etc/skitaid/servers-enabled")
		mkdir ("/etc/skitaid/servers-available")
		mkdir ("/var/log/skitaid")
		mkdir ("/var/local/skitaid")		
		mkdir ("/var/local/skitaid-pub/sample/static")
		mkdir ("/var/local/skitaid-pub/sample/templates")
			
		conf = "/etc/skitaid/skitaid.conf"
		if not os.path.isfile (conf) and not os.path.islink (conf):
			shutil.copyfile (os.path.join ("skitai/skitaid/etc/skitaid/skitaid.conf"), "/etc/skitaid/skitaid.conf")
		
		with open ("skitai/skitaid/etc/skitaid/servers-enabled/sample.conf") as f:
			data = f.read ().replace ("c:\\skitaid\\pub\sample\\", "/var/local/skitaid-pub/sample/")
		with open ("/etc/skitaid/servers-enabled/sample.conf", "w") as f:
			f.write (data)
		
		for path in ("cert/README.txt", "servers-available/README.txt"):
			try: 
				os.remove (os.path.join ("/etc/skitaid", path))
			except OSError as why:
				if why.errno == 2: pass
			shutil.copyfile (os.path.join ("skitai/skitaid/etc/skitaid", path), os.path.join ("/etc/skitaid", path))
		
		for fn in ("skitaid.py", "skitaid-instance.py", "skitaid-smtpda.py", "skitaid-cron.py"):
			target = os.path.join ("/usr/local/bin", fn)			
			exists_python = None
			if os.path.isfile (target):
				with open (target, "rb") as f:
					exists_python = f.readline ().strip ()
					if not exists_python.startswith (b"#!"):
						exists_python = None
			
			with open (os.path.join ("skitai/skitaid/bin", fn), "rb") as f:
				data = f.read ()				
			with open (target, "wb") as f:			
				if exists_python:
					data = data.replace (b"#!/usr/bin/python", exists_python, 1)
				f.write (data)
			os.chmod (target, 0o755)
		
		for fn in ("static/index.html",):
			target = os.path.join ("/var/local/skitaid-pub/sample", fn)
			try: os.remove (target)
			except OSError as why:
				if why.errno == 2: pass
					
		for fn in ("webapp.py", "templates/documentation.html", "templates/websocket.html", "templates/index.html", "static/reindeer.jpg"):
			target = os.path.join ("/var/local/skitaid-pub/sample", fn)
			try: os.remove (target)
			except OSError as why:
				if why.errno == 2: pass
			shutil.copyfile (os.path.join ("skitai/skitaid/pub/sample", fn), target)
		
		if os.path.isfile ("/etc/init.d/skitaid"):
			os.remove ("/etc/init.d/skitaid")
		shutil.copyfile ("skitai/skitaid/etc/init.d/skitaid", "/etc/init.d/skitaid")
		os.chmod ("/etc/init.d/skitaid", 0o755)

		print("\n\n======================================")
		print("Installation Complete")
		print("--------------------------------------")	
		print("Please run below commands\n")
		print("  sudo service skitaid start")
		print("  wget http://localhost:5000\n\n")
		



"""
This module is copy of skitai.handler.pingpong_handler.py.
It's not actually used by skitai, but for just your reference
"""

from . import wsgi_handler

class Handler (wsgi_handler.Handler):
	def __init__(self, wasc):
		self.wasc = wasc		
		
	def match (self, request):
		return request.split_uri() [0] == "/ping"
		
	def handle_request (self, request):
		# WSGI Server Payload Emulating
		env = self.build_environ (request)
		start_response = request.response.start_response
		
		start_response ("200 OK", [("Content-Type", "text/plain"), ("Content-Length", "4")])
		request.response.push (b"pong")
		request.response.done ()
		